#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ________________ HEADER _________________________

"""skeletoncdf module.

Program to convert a CDF skeleton table into
a binary CDF ("master").

Skeleton/binary CDF can also be generated from
a formatted Excel skt_editor file.
"""

# ________________ IMPORT _________________________
# (Include here the modules to import, e.g. import sys)
import os
import os.path as osp
import logging

from maser.tools.toolbox import which, run_command

from maser.tools.cdf.serializer.skeleton import Skeleton

# ________________ HEADER _________________________

__all__ = ["skeletoncdf"]

# ________________ Global Variables _____________
# (define here the global variables)
logger = logging.getLogger(__name__)

# ________________ Class Definition __________
# (If required, define here classes)


def skeletoncdf(
    input_skt,
    output_dir=os.getcwd(),
    output_cdf=None,
    overwrite=False,
    auto_pad=False,
    exe=None,
    no_cdf=False,
):
    """
    Make a CDF Master binary file from a ASCII
    skeleton table using the skeletoncdf program of the NASA CDF software.

    If the "excel_format" keyword is True, then
    convert first the input Excel skeleton file into a valid CDF skeleton
    table.


    :param input_skt: Path of the input skeleton table file
    :param output_dir: Path of the output directory
    :param overwrite: If True, overwrite existing output file
    :param auto_pad: Automatically define !VAR_PADVALUE if set to True
    :param exe: Path of the skeletoncdf program executable
    :param no_cdf: If True, generate the skeleton table only. Only works with excel_format keyword
    :return:
    """
    # If output_dir does not provide then use current one
    # If provided, but does not exist, then create it
    if not osp.isdir(output_dir):
        logger.warning("{0} output directory not found, create it!".format(output_dir))
        os.mkdir(output_dir)

    # If the input file is an Excel 2007 format file...
    basename, extension = os.path.splitext(input_skt)

    if extension == ".xlsx":

        input_xlsx = input_skt
        logger.info(
            "First converting {0} into valid skeleton table...".format(input_xlsx)
        )
        skeleton = Skeleton.from_xlsx(input_skt, auto_pad=auto_pad)
        input_skt = skeleton.to_txt(output_path=output_dir, overwrite=overwrite)

        # If conversion has failed...
        if input_skt is None:
            input_skt = os.path.splitext(input_xlsx)[0] + ".skt"
            logger.error('OUTPUT "{0}" HAS NOT BEEN SAVED!'.format(input_skt))
            return None

    if no_cdf:
        return input_skt

    # Set output_cdf file path
    if output_cdf is None:
        output_cdf = osp.basename(osp.splitext(input_skt)[0]) + ".cdf"

    output_cdf = osp.join(output_dir, os.path.basename(output_cdf))

    # Initialize command line
    cmd = []

    # If skeletoncdf program path is not provided
    # then search it on the $PATH
    if exe is None:
        if "CDF_BIN" in os.environ:
            exe = osp.join(os.environ["CDF_BIN"], "skeletoncdf")
        else:
            exe = which("skeletoncdf")
    if exe is None:
        logger.error("skeletoncdf program is not callable!")
        return None
    cmd.append(exe)
    if os.path.isfile(output_cdf) and overwrite:
        logger.warning("%s existing file will be overwritten!", output_cdf)
        cmd.append("-delete")
    cmd.append(input_skt)
    cmd.extend(["-cdf", output_cdf])
    myenv = os.environ.copy()
    logger.info("Executing {0}...".format(" ".join(cmd)))
    res = run_command(cmd, env=myenv)
    output, errors = res.communicate()
    if res.wait() == 0:
        logger.debug(output)
        if os.path.isfile(output_cdf):
            logger.info(output_cdf + " has been saved correctly!")
            return output_cdf
        else:
            logger.error(output_cdf + " has not been saved correctly!")
    else:
        logger.error("ERROR RUNNING COMMAND: ")
        logger.error(" ".join(cmd))
        logger.error("STDOUT - %s", str(output))
        logger.error("STDERR - %s", str(errors))
        logger.error('OUTPUT "{0}" HAS NOT BEEN SAVED!'.format(output_cdf))

    return None


# _________________ Main ____________________________
if __name__ == "__main__":
    print(__file__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""
Contains exceptions for the cdf module.

"""

import logging

logger = logging.getLogger(__name__)

__all__ = ["InvalidFile", "InvalidSkeleton", "InvalidEntry", "CDFSerializerError"]


class InvalidFile(Exception):
    """Exception raise if the input file is invalid."""

    def __init__(self, message="", *args, **kwargs):
        super(InvalidFile, self).__init__(*args, **kwargs)
        logger.error(message)
        self.message = message


class InvalidSkeleton(Exception):
    pass


class InvalidEntry(Exception):
    def __init__(self, message="", *args, **kwargs):
        super(InvalidEntry, self).__init__(*args, **kwargs)
        logger.error(message)
        self.message = message


class CDFSerializerError(Exception):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""Python 3 module to validate a CDF format file."""

# ________________ IMPORT _________________________
# (Include here the modules to import, e.g. import sys)
import os
import json
import logging
from tempfile import TemporaryDirectory

from spacepy.pycdf import CDF, zAttr

from maser.tools.toolbox import run_command, quote, move_safe
from maser.tools.cdf.tools import get_cdftype, get_cdftypename
from maser.tools.settings import SUPPORT_DIR

__all__ = ["Validate", "cdfvalidator", "ValidatorException"]

# ________________ HEADER _________________________


# ________________ Global Variables _____________
# (define here the global variables)
logger = logging.getLogger(__name__)

CDF_ENV = {"CDF_LEAPSECONDSTABLE": None, "CDF_BIN": None}

ISTP_MOD_FILE = os.path.join(SUPPORT_DIR, "tools", "validator_model_istp.json")

# Possible values for Issue class types
ISSUE_TYPES = ["zvar", "gatt", "vatt"]

# Possible values for Issue class checks
ISSUE_CHECKS = [
    "isitem",
    "hasvalue",
    "isentry",
    "istype",
    "issize",
    "isdims",
    "isattrs",
]


# ________________ Class Definition __________
# (If required, define here classes)
class ValidatorException(Exception):
    """Exception class for Validator."""

    pass


class Issue:
    """Issue class."""

    def __init__(self, issues=None):
        self.counter = 0
        self.reset()
        if issues:
            self.extend(issues)

    def append(self, name, itype, check, msg, passed, varname="", num=None):
        """
        Append a new issue element

        :param name: Name of the issue
        :param itype: type of the issue
        :param check: check method of the issue
        :param msg: resulting message of the issue
        :param passed: if issue has succeeded is True, else False
        :param varname: Optional attribute for zvar name (in case vatt issue)
        :param num: gives issue number (if not provided, use counter)
        :return:
        """

        if itype not in ISSUE_TYPES:
            logger.warning("Input Issue type ({0}) is not valid!".format(itype))
            return None

        if check not in ISSUE_CHECKS:
            logger.warning("Input Issue check ({0}) is not valid!".format(check))
            print(name, itype, check, msg, passed)
            _ = input()
            return None

        if not num:
            self.num.append(self.counter)
        else:
            self.num.append(num)
        self.name.append(name)
        self.type.append(itype)
        self.msg.append(msg)
        self.passed.append(passed)
        self.check.append(check)
        self.varname.append(varname)
        self.counter += 1

    def extend(self, issues):
        """
        Add issues to Issue object

        :param issues:
        :return:
        """

        if Issue.is_valid(issues):
            for issue in issues:
                self.append(issue[1], issue[2], issue[3], issue[4], issue[5], issue[6])
        else:
            logger.warning("Input argument is not valid Issue object!")

    @staticmethod
    def is_valid(issue):
        """
        Check if input is a valid Issue object

        :param issue: input to verify
        :return: True if is input is a valid Issue object, False otherwise
        """
        return isinstance(issue, Issue)

    def is_passed(self):
        """
        Check if issues are successfully passed or not

        :return: return True if all issues have been passed successfully, else False
        """
        return False not in self.passed

    def __len__(self):
        """Return number of issue elements."""
        return len(self.num)

    def __str__(self):
        """Return string with list of issues"""
        string = ""
        for issue in self:
            string += str(issue) + "\n"
        return string

    def reset(self):
        """
        Reset Issue object.

        :return:
        """
        self.num = []
        self.name = []
        self.type = []
        self.check = []
        self.msg = []
        self.passed = []
        self.varname = []
        self.counter = 0

    def __iter__(self, issues=None):
        """Iterator for Issue class."""

        if not issues:
            issues = self

        for i in range(0, len(issues)):
            yield issues.num[i], issues.name[i], issues.type[i], issues.check[
                i
            ], issues.msg[i], issues.passed[i], issues.varname[i]

    def to_dict(self):
        """
        Convert issues into dictionary.

        :return:
        """
        # Generate dictionary from Issue object
        issue_dict = dict()
        for issue in self:
            issue_dict[issue[0]] = {
                "name": issue[1],
                "type": issue[2],
                "check": issue[3],
                "msg": issue[4],
                "passed": issue[5],
            }
            if issue[6]:
                issue_dict[issue[0]]["varname"] = issue[6]
        return issue_dict

    def to_json(self, output_file, overwrite=False, comment=""):
        """
        Write issues into an output json format file.

        :param output_file: Name of the output JSON file
        :param overwrite: If True then overwrite existing file
        :return: Issue dictionary
        """

        if os.path.isfile(output_file):
            if overwrite:
                os.remove(output_file)
            else:
                logger.warning("{0} already exists, aborting!".format(output_file))
                return None

        issue_dict = self.to_dict()
        with open(output_file, "w") as fw:
            json.dump(issue_dict, fw)

        if os.path.isfile(output_file):
            logger.info("{0} saved".format(output_file))
            return issue_dict
        else:
            logger.warning("Saving {0} has failed!".format(output_file))
            return None


class Validate:
    """Class that provides tools to validate a CDF format file."""

    def __init__(self, cdf_file, cdf_env=CDF_ENV):
        """Validate Init method."""

        self.file = None
        self.cdf = None
        self.open_cdf(cdf_file)
        self.cdf_env = self._init_cdfenv(cdf_env=cdf_env)
        self.issues = Issue()

    def _init_cdfenv(self, cdf_env=CDF_ENV):
        """Initialize instance."""
        # get CDF program path
        for key, val in cdf_env.items():
            if val is None:
                if key in os.environ:
                    cdf_env[key] = os.environ[key]
                else:
                    logger.error("{0} is not defined!".format(key))
                    raise ValidatorException
        return cdf_env

    def open_cdf(self, file):
        """
        Open the input cdf file.

        :param file:
        :return:
        """
        logger.info("Opening {0}".format(file))
        self.file = file
        try:
            cdf = CDF(self.file)
            cdf.readonly(True)
        except ValidatorException as e:
            logger.error(e)
            raise ValidatorException
        else:
            self.cdf = cdf

    def close_cdf(self):
        """Close current cdf file."""
        logger.info("Closing " + self.file)
        self.cdf.close()

    def is_istp_compliant(self):
        """
        Istp compliant.

        Check that the input CDF is compliant with
        ISTP guidelines

        :return:
        """
        issues = Issue()

        issues.extend(self.is_model_compliant(ISTP_MOD_FILE))
        issues.extend(self.is_istp_fillval())

        self.issues.extend(issues)

        return issues

    def is_model_compliant(self, model_file):
        """
        Check the CDF content against a model given as an input JSON file.

        :param model_file: input json file path
        :return:
        """
        pass

        def check(cdf, items, is_vattr=None, is_gattr=None, varname=""):
            """
            Check items of a given CDF
            GLOBALattributes, VARIABLEattributes, zVariables

            :param cdf: CDF object to check
            :param items: List of CDF items to check
            :param vattr: Check variable attributes
            :param gattr: Check global attributes
            :return:
            """
            issues = Issue()

            for item in items:
                name = item["name"]
                logger.info('Checking "{0}"'.format(name))
                istype = "type" in item
                isentry = "entries" in item
                hasvalue = "hasvalue" in item
                isexcluded = "excludes" in item

                issizes = "sizes" in item
                isdims = "dims" in item
                isattrs = "attributes" in item

                if isexcluded and varname in item["excludes"]:
                    logger.info("Checking {0} for {1} is ignored".format(name, varname))
                    return issues

                if is_vattr:
                    issue_type = "vatt"
                    try:
                        entries = self.cdf[varname].attrs
                    except Exception:
                        entries = []
                elif is_gattr:
                    issue_type = "gatt"
                    entries = self.cdf.attrs
                else:
                    issue_type = "zvar"
                    entries = cdf

                if name in entries:
                    cdfitem = entries[name]

                    # If variable attribute...
                    if is_vattr:
                        # Get corresponding zAttr object
                        zattr = zAttr(cdf, name)

                        # Get number of entries
                        nentry = len(zattr)
                        # Get data type (from the max_idx to make sure)
                        dtype = get_cdftypename(zattr.type(zattr.max_idx()))
                    # If global attribute...
                    elif is_gattr:
                        # Get number of entries
                        nentry = len(cdfitem)

                        # Get data type (from the max_idx to make sure)
                        dtype = get_cdftypename(cdfitem.type(cdfitem.max_idx()))
                    # If zvariable...
                    else:
                        # If zVariable, get data type ...
                        dtype = get_cdftypename(cdfitem.type())
                        # and entry number
                        nentry = len(cdfitem)

                    # Store the item existence checking
                    passed = True
                    msg = "{0} CDF item found".format(name)
                    logfunc = logger.info
                    issues.append(
                        name, issue_type, "isitem", msg, passed, varname=varname
                    )
                    logfunc(msg)

                    # Check if item has a value
                    if hasvalue:
                        logger.info("Checking value existence...")
                        if nentry == 0:
                            msg = '"{0}" has no entry value!'.format(name)
                            logfunc = logger.warning
                            passed = False
                        else:
                            msg = '"{0}" has entry values'.format(name)
                            logfunc = logger.info
                            passed = True
                        issues.append(
                            name, issue_type, "hasvalue", msg, passed, varname=varname
                        )
                        logfunc(msg)

                    if istype:
                        logger.info("Checking data type")
                        cdftype = get_cdftypename(get_cdftype(item["type"]))
                        if dtype != cdftype:
                            msg = '"{0}" has the wrong data type: '.format(
                                name
                            ) + '"{0}" found, but "{1}" expected!'.format(
                                dtype, cdftype
                            )
                            logfunc = logger.warning
                            passed = False
                        else:
                            passed = True
                            logfunc = logger.info
                            msg = '"{0}" data type found'.format(dtype)

                        issues.append(
                            name, issue_type, "istype", msg, passed, varname=varname
                        )
                        logfunc(msg)

                    if issizes:
                        logger.info("Checking data size")
                        if list(cdfitem.shape[1:]) != list(item["sizes"]):
                            msg = '"{0}" has the wrong sizes: '.format(
                                name
                            ) + '"{0}" found, but "{1}" expected!'.format(
                                cdfitem.shape[1:], item["sizes"]
                            )
                            passed = False
                            logfunc = logger.warning
                        else:
                            passed = True
                            msg = '"{0}" sizes found'.format(item["sizes"])
                            logfunc = logger.info
                        issues.append(
                            name, issue_type, "issize", msg, passed, varname=varname
                        )
                        logfunc(msg)

                    if isdims:
                        logger.info("Checking data dimension(s)")
                        if len(cdfitem.dv()) != item["dims"]:
                            msg = '"{0}" has the wrong dims: '.format(
                                name
                            ) + '"{0}" found, but "{1}" expected!'.format(
                                len(cdfitem), item["dims"]
                            )
                            logfunc = logger.warning
                            passed = False
                        else:
                            passed = True
                            msg = '"{0}" dims found!'.format(item["dims"])
                            logfunc = logger.info
                        issues.append(
                            name, issue_type, "isdims", msg, passed, varname=varname
                        )
                        logfunc(msg)

                    # Check if item has entries
                    if isentry:
                        logger.info("Checking attribute entries...")
                        if nentry < len(item["entries"]):
                            msg = '"{0}" has missing entries!'.format(name)
                            logger.warning(msg)
                            issues.append(
                                name, issue_type, "isentry", msg, False, varname=varname
                            )
                        else:
                            # To avoid error in case where cdfitem is a String (for vattr)
                            if isinstance(cdfitem, str):
                                cdfitem = [cdfitem]
                            for i, entry in enumerate(item["entries"]):
                                if cdfitem[i] != entry:
                                    msg = '"{0}" has a wrong entry value: '.format(
                                        name
                                    ) + '"{0}" found, but "{1}" expected!'.format(
                                        cdfitem[i], entry
                                    )
                                    logfunc = logger.warning
                                    passed = False
                                else:
                                    passed = True
                                    msg = '"{0}" atttribute entry found'.format(entry)
                                    logfunc = logger.info
                                issues.append(
                                    name,
                                    issue_type,
                                    "isentry",
                                    msg,
                                    passed,
                                    varname=varname,
                                )
                                logfunc(msg)

                    if isattrs:
                        logger.info(
                            'Checking variable attributes of "{0}"...'.format(name)
                        )
                        check(cdf, item["attributes"], is_vattr=True, varname=name)
                else:
                    msg = '"{0}" required!'.format(name)
                    logging.warning(msg)
                    passed = False
                    issues.append(
                        name, issue_type, "isitem", msg, passed, varname=varname
                    )

            return issues

        issues = Issue()

        # Read JSON format model file
        logger.info("Loading " + model_file)
        with open(model_file, "r") as fbuff:
            mfile = json.load(fbuff)

        if "GLOBALattributes" in mfile:
            logging.info("Checking GLOBALattributes:")
            issues.extend(check(self.cdf, mfile["GLOBALattributes"], is_gattr=True))
        if "VARIABLEattributes" in mfile:
            # vattrs = get_vattrs(cdf)
            # Loop over all zVariables
            for zvar in self.cdf.keys():
                logging.info("Checking VARIABLEattributes for {0}".format(zvar))
                issues.extend(
                    check(
                        self.cdf,
                        mfile["VARIABLEattributes"],
                        is_vattr=True,
                        varname=zvar,
                    )
                )
        if "zVariables" in mfile:
            logging.info("Checking zVariables:")
            issues.extend(check(self.cdf, mfile["zVariables"]))

        return issues

    def is_istp_fillval(self, zvarnames=None):
        """
        Is_istp_fillval.

        Check if the FILLVAL variable attribute values
        are ISTP compliant
        """
        issues = Issue()

        if zvarnames is None:
            zvarnames = [key for key in self.cdf]

        # Read JSON format model file
        logger.info("Loading " + ISTP_MOD_FILE)
        with open(ISTP_MOD_FILE, "r") as fbuff:
            mfile = json.load(fbuff)

        istpfillval = mfile["ISTPMapping"]["FILLVAL"]

        for zvname in zvarnames:
            if zvname in self.cdf:
                zvar = self.cdf[zvname]
                if "FILLVAL" in zvar.attrs:
                    fillval = zvar.attrs["FILLVAL"]
                    zvtype = get_cdftypename(zvar.type())
                    if str(istpfillval[zvtype]) != str(fillval):
                        msg = "%s has an invalid FILLVAL value: " % (quote(zvname))
                        msg += "%s found, but %s expected!" % (
                            quote(fillval),
                            quote(istpfillval[zvtype]),
                        )
                        logging.warning(msg)
                        issues.append(
                            "FILLVAL", "vatt", "isentry", msg, False, varname=zvar
                        )
                    else:
                        msg = "FILLVAL found in {0}".format(zvname)
                        issues.append(
                            "FILLVAL", "vatt", "isentry", msg, True, varname=zvar
                        )
                else:
                    msg = "%s has no FILLVAL attribute!" % (quote(zvname))
                    logging.warning(msg)
                    issues.append(
                        "FILLVAL", "vatt", "isentry", msg, False, varname=zvar
                    )
            else:
                logger.warning("%s not found in %s!", quote(zvar), self.cdf_file)

        return issues

    def cdfconvert(self, src, dst, args=None, program=None, overwrite=False):
        """Run the cdfconvert program in the NASA CDF distribution."""
        cmd = []

        if program is None:
            program = os.path.join(self.cdf_env["CDF_BIN"], "cdfconvert")

        if program is None or not os.path.isfile(program):
            logger.error("cdfconvert not found!")
            return None

        cmd = [program, src, dst]

        if args is not None:
            cmd.append(args)

        if overwrite is True:
            with TemporaryDirectory() as tempdir:
                src = move_safe(src, tempdir)
                self.cdfconvert(src, dst, args=args, program=program, overwrite=False)

        return run_command(cmd)

    def cdfvalidate(self, file, program=None):
        """Run the cdfvalidate program in the NASA CDF distribution."""
        cmd = []

        if program is None:
            program = os.path.join(self.cdf_env["CDF_BIN"], "cdfvalidate")

        if program is None or not os.path.isfile(program):
            logger.error("cdfvalidate not found!")
            return None
        cmd = [program, file]

        return run_command(cmd)

    @classmethod
    def is_cdf_valid(cls, file, program=None, quiet=False):
        """Call the cdfvalidate program."""
        res = cls(quiet=quiet).cdfvalidate(file, program=program)

        output, errors = res.communicate()
        if res.wait() == 0:
            logger.info(output)
            return True
        else:
            logger.error("ERROR RUNNING COMMAND cdfvalidate: ")
            logger.error("STDOUT - %s", str(output))
            logger.error("STDERR - %s", str(errors))
            return False

    def is_zvar_valid(self, zvarname):
        """
        Is_zvar_valid.

        Check  the values of a zVariable
        comparing to the its VALIMIN/VALIDMAX attributes
        """
        issues = Issue()

        cdf = self.cdf

        if zvarname not in cdf:
            logger.error("%s not in %s!", zvarname, cdf)
            return False
        else:
            zvar = cdf[zvarname]

        logger.info("Cheching %s", zvarname)

        zattrs = zvar.attrs
        if "VALIDMIN" in zattrs:
            validmin = zattrs["VALIDMIN"]
            logger.info("VALIDMIN=%s", str(validmin))
        else:
            logger.warning("No VALIDMIN attribute for %s", zvarname)
            validmin = None

        if "VALIDMAX" in zattrs:
            validmax = zattrs["VALIDMAX"]
            logger.info("VALIDMAX=%s", str(validmax))
        else:
            logger.warning("No VALIDMAX attribute for %s", zvarname)
            validmax = None

        for i, rec in enumerate(zvar):
            if rec.min() < validmin:
                msg = ("[%i]: Record value(s) lesser than VALIDMIN!") % (i)
                logger.warning(msg)
                issues.append(
                    "VALIDMIN", "vatt", "isentry", msg, False, varname=zvarname
                )

            if rec.max() > validmax:
                msg = ("[%i]: Record value(s) greater than VALIDMAX!") % (i)
                logger.warning(msg)
                issues.append(
                    "VALIDMAX", "vatt", "isentry", msg, False, varname=zvarname
                )

        return issues

    def flush_issues(self):
        """
        Reset issue object

        :return:
        """
        self.issues.reset()


# ________________ Global Functions __________
# (If required, define here gobal functions)
def cdfvalidator(
    cdf_file,
    is_istp=False,
    model_json_file=None,
    cdfvalidate_bin=None,
    run_cdf_validate=False,
):
    """cdfvalidator main program."""

    # Initialize a Validate object
    cdfvalid = Validate(cdf_file=cdf_file)

    # Check ISTP compliance
    if is_istp:
        cdfvalid.is_istp_compliant()

    # Check compliance with input model json file
    if model_json_file:
        cdfvalid.is_model_compliant(model_json_file)

    cdfvalid.close_cdf()


# _________________ Main ____________________________
if __name__ == "__main__":
    print(__file__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ________________ HEADER _________________________

"""skeletontable module.

Program to convert a CDF file into
a skeleton table.

Skeleton table can also be saved into a Excel 2007 format file.
"""

# ________________ IMPORT _________________________
# (Include here the modules to import, e.g. import sys)
import os
import os.path as osp
import logging

from maser.tools.toolbox import which, run_command

from maser.tools.cdf.serializer.skeleton import Skeleton

# ________________ HEADER _________________________

__all__ = ["skeletontable"]

# ________________ Global Variables _____________
# (define here the global variables)
logger = logging.getLogger(__name__)

# ________________ Class Definition __________
# (If required, define here classes)


def skeletontable(
    input_cdf,
    output_dir=os.getcwd(),
    output_skt=None,
    overwrite=False,
    to_xlsx=False,
    exe=None,
):
    """
    Make a skeleton table file from an input CDF file
    using the skeletontable program of the NASA CDF software.

    If the "excel_format" keyword is True, then
    the skeleton table is also saved into an Excel 2007 format file.

    :param input_cdf: Input CDF file to convert
    :param output_dir: Path of the output directory
    :param output_skt: Output skeleton file
    :param overwrite: If True, overwrite existing output file
    :param to_xlsx: If True, also save the skeleton table in an Excel 2007 format file
    :param exe: Path to the "skeletontable" executable
    :return:
    """

    # If output_dir does not provide then use current one
    # If provided, but does not exist, then create it
    if not osp.isdir(output_dir):
        logger.warning("{0} output directory not found, create it!".format(output_dir))
        os.mkdir(output_dir)

    # Get basename, extension of the input CDF
    basename, extension = os.path.splitext(input_cdf)

    # Set output_skt file path
    if output_skt is None:
        output_skt = osp.basename(basename) + ".skt"

    output_skt = osp.join(output_dir, os.path.basename(output_skt))

    # If input file is already a skeleton then skip this step
    if extension != ".skt":

        # Initialize command line
        cmd = []

        # If skeletontable program path is not provided
        # then search it on the $PATH
        if exe is None:
            if "CDF_BIN" in os.environ:
                exe = osp.join(os.environ["CDF_BIN"], "skeletontable")
            else:
                exe = which("skeletontable")
        if exe is None:
            logger.error("skeletontable program is not callable!")
            return None
        cmd.append(exe)
        if os.path.isfile(output_skt) and overwrite:
            logger.warning("%s existing file will be overwritten!", output_skt)
            cmd.append("-delete")
        cmd.append(input_cdf)
        cmd.extend(["-skeleton", os.path.splitext(output_skt)[0]])
        myenv = os.environ.copy()
        logger.info("Executing {0}...".format(" ".join(cmd)))
        res = run_command(cmd, env=myenv)
        output, errors = res.communicate()
        if res.wait() == 0:
            logger.debug(output)
            if os.path.isfile(output_skt):
                logger.info("{0} has been saved correctly!".format(output_skt))
                input_skt = output_skt
            else:
                logger.error("{0} has not been saved correctly!".format(output_skt))
                return None
        else:
            logger.error("ERROR RUNNING COMMAND: ")
            logger.error(" ".join(cmd))
            logger.error("STDOUT - %s", str(output))
            logger.error("STDERR - %s", str(errors))
            logger.error('OUTPUT "{0}" HAS NOT BEEN SAVED!'.format(output_skt))
            return None
    else:
        logger.warning(
            "Input CDF seems to be already a skeleton table ({0})".format(input_cdf)
        )
        input_skt = input_cdf

    # if requested, also save as an Excel file
    if to_xlsx:
        skeleton = Skeleton.from_txt(input_skt)
        output_xlsx = os.path.join(
            output_dir, os.path.basename(os.path.splitext(input_skt)[0] + ".xlsx")
        )
        skeleton.to_xlsx(output_xlsx)
        if os.path.isfile(output_xlsx):
            logger.info("{0} has been saved correctly!".format(output_xlsx))
        else:
            logger.error("{0} has not been saved correctly!".format(output_xlsx))

    return output_skt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ________________ HEADER _________________________


"""cdf.serializer.txt module

Contains the class to manage the Skeleton table in text format.

"""


# ________________ IMPORT _________________________
# (Include here the modules to import, e.g. import sys)
import os
import re
import logging
from datetime import datetime
from copy import deepcopy

from jinja2 import Environment, FileSystemLoader

from maser.tools.settings import __version__
from maser.tools.cdf.serializer.exceptions import InvalidFile
from maser.tools.cdf.serializer.globals import JINJA_TEMPLATE_DIR
from maser.tools.cdf.serializer.globals import (
    SHEETS,
    HEADER,
    GATTRS,
    ZVARS,
    RVARS,
    VATTRS,
    NRV,
    END,
)
from maser.tools.cdf.serializer.globals import (
    NEW_HEADER,
    NEW_ZVAR,
    NEW_VATTRS,
    NO_NRV,
    NEW_NRV,
)


__all__ = ["Skt2txt", "Txt2skt"]

# ________________ Global Variables _____________
# (define here the global variables)
logger = logging.getLogger(__name__)


SKT_TEMPLATE = "skeleton.skt-tpl"

IN_FTIME = "%Y-%m-%dT%H:%M:%S"
NOW = datetime.now()
ROW_LENGTH_MAX = 79
DEF_INDENT = " " * 16


# ________________ Class Definition __________
# (If required, define here classes)
class Skt2txt:
    """Class to convert a Skeleton object into a CDF skeleton table."""

    def __init__(self, skeleton, ignore_none=True, auto_pad=True):
        """__init__ method."""
        self.skeleton = deepcopy(skeleton)
        self.render = None
        self.ignore = ignore_none
        self.auto_pad = auto_pad

    def _build_render(self):
        """Build the CDF skeleton table content using the Skeleton object."""
        logger.debug("Building CDF skeleton table body... ")

        file = self.skeleton.file
        skt_name = os.path.splitext(os.path.basename(file))[0]

        # Setup jinja2 environment
        jenv = Environment(loader=FileSystemLoader(str(JINJA_TEMPLATE_DIR)))

        # Load skeleton table template
        template = jenv.get_template(str(SKT_TEMPLATE))

        # Build the Skeleton template render
        self.render = template.render(
            gen_time=NOW.strftime(IN_FTIME),
            name=skt_name,
            version=__version__,
            file=os.path.basename(file),
            header=self.skeleton.header,
            gattrs=self.reform_gattr(self.skeleton.gattrs),
            zvars=self.skeleton.zvars,
            vattrs=self.reform_vattr(self.skeleton.vattrs),
            vattrList=self.skeleton.vattrList,
        )

    def reform_gattr(self, gattrs, length=48):
        """
        Re-format the global attributess to fit inside skeleton table
        (i.e., break too long line.)

        :param gattrs: list of g. attributes
        :return:
        """

        for gattr, entries in gattrs.items():
            for i, entry in enumerate(entries):
                value = ""
                counter = 0
                if str(entry["Value"]) == "None":
                    value = " "
                elif entry["Data Type"] != "CDF_CHAR":
                    continue
                else:
                    for char in str(entry["Value"]):
                        value += char
                        counter += 1
                        if counter > length:
                            value += '" - \n "'
                            counter = 0

                gattrs[gattr][i]["Value"] = value

        return gattrs

    def reform_vattr(self, vattrs, length=60):
        """
        Re-format the variable attributess to fit inside skeleton table
        (i.e., break too long line.)

        :param vattrs: list of v. attributes
        :return:
        """

        for zvar, entries in vattrs.items():
            for vatt, entry in entries.items():
                value = ""
                counter = 0
                if str(entry["Value"]) == "None":
                    value = " "
                elif entry["Data Type"] != "CDF_CHAR":
                    continue
                else:
                    for char in entry["Value"]:
                        value += char
                        counter += 1
                        if counter > length:
                            value += '" - \n "'
                            counter = 0

                vattrs[zvar][vatt]["Value"] = value

        return vattrs

    def write_txt(self, output_path=None, overwrite=False):
        """Write the CDF skeleton table file."""

        if not output_path:
            txt = os.path.splitext(self.skeleton.file)[0] + ".skt"
        elif os.path.isdir(output_path):
            txt = os.path.join(
                output_path,
                os.path.basename(os.path.splitext(self.skeleton.file)[0] + ".skt"),
            )
        else:
            txt = os.path.splitext(output_path)[0] + ".skt"

        if not (overwrite) and (os.path.isfile(txt)):
            logger.warning("%s already exits!", txt)
            return txt

        logger.info("Writing %s...", txt)

        if not self.render:
            logger.debug("Attempting to build Skeleton table body...")
            self._build_render()

        with open(txt, "w") as filew:
            filew.write(self.render)

        if os.path.isfile(txt):
            logger.info("{0} has been saved correctly".format(txt))
            return txt
        else:
            logger.error("{0} has not been saved correctly!".format(txt))
            return None


class Txt2skt:
    """Class to convert a CDF skeleton table file into a Skeleton object."""

    def __init__(self, skeleton):

        self.file = None
        self.skeleton = skeleton
        self.section = None
        self.row = ""

    def parse_txt(self, txt_file):
        """
        Parse input CDF skeleton table file

        :param txt_file: Path to CDF skeleton table file
        :return:
        """
        self.file = txt_file
        if not os.path.isfile(txt_file):
            logger.error("{0} NOT FOUND, ABORTING!".format(txt_file))
            raise FileNotFoundError
        else:
            self.file = txt_file

        zvar = None
        new_header = False
        new_zvar = False
        new_vattrs = False
        new_nrv = False
        gattname = None
        for row in self.txt_iterator():
            cols = [col for col in re.split(r"\[|\]| |'|\n", row) if col]
            ncols = len(cols)
            # Get "compressed" row, i.e., without any empty space
            row_c = "".join(cols)

            logger.debug(row_c)

            # Enter into the header section
            if self.section == HEADER:

                if row_c == NEW_HEADER:
                    new_header = True
                    continue
                elif row_c.startswith("!-"):
                    continue
                elif new_header:
                    self.skeleton.header["nzvar"] = cols[0].split("/")[1]
                    self.skeleton.header["ngattr"] = cols[1]
                    self.skeleton.header["nvattr"] = cols[2]
                    new_header = False
                    continue

                for field in SHEETS[HEADER]:
                    if field in row:
                        self.skeleton.header[field] = row.strip().split(":")[1].strip()
                        break
                    elif row.startswith("!") or ncols == 0:
                        continue

            # Enter into the global attribute definition section
            elif self.section == GATTRS:
                # Skip comment/empty line
                if row_c.startswith("!") or ncols == 0:
                    continue
                else:
                    # If it is a global attribute definition row...
                    self.row += " " + row

                if row_c.endswith("}") or row_c.endswith("}."):
                    # If gattname already defined, it means that it is a new entry
                    # Then append the name of g. attribute for the field
                    # extraction
                    if gattname:
                        self.row = '"{0}" '.format(gattname) + self.row
                    # Extract field for the current g.attribute entry
                    gatt_desc = self._extract_attr()
                    # Store the g.attribute fields into self.skeleton.gattrs
                    if gatt_desc["Attribute Name"] in self.skeleton.gattrs:
                        self.skeleton.gattrs[gatt_desc["Attribute Name"]].append(
                            gatt_desc
                        )
                    else:
                        self.skeleton.gattrs[gatt_desc["Attribute Name"]] = [gatt_desc]

                    # Define gatt_desc["Attribute Name"] as the current
                    # gattname
                    gattname = gatt_desc["Attribute Name"]

                    self.row = ""

                # If it is the end of the global attribute definition,
                # Reset the gattname
                if row_c.endswith("."):
                    gattname = None
                    self.row = ""

            # Enter into the variable attribute list section
            elif self.section == VATTRS:
                # Skip comment/empty line
                if row.startswith("!") or row == "":
                    continue
                else:
                    self.skeleton.vattrList.append(row.split('"')[1])

            # Enter into the zVariable definition section
            elif self.section == ZVARS:

                # Retrieve value of the VAR_COMPRESSION parameter
                if "!VAR_COMPRESSION:" in row_c:
                    value = row.strip().split(":")
                    nval = len(value)
                    if nval == 2:
                        value = value[1]
                    elif nval > 2:
                        value = ":".join(value[1:])
                    else:
                        logger.error("VAR_COMPRESSION seems to be badly formatted!")
                        raise InvalidFile

                    self.skeleton.zvars[zvar]["VAR_COMPRESSION"] = value
                # Retrieve value of the VAR_PADVALUE parameter
                elif "!VAR_PADVALUE:" in row_c:
                    value = row.strip().split(":")
                    nval = len(value)
                    if nval == 2:
                        value = value[1]
                    elif nval > 2:
                        value = ":".join(value[1:])
                    else:
                        logger.error("VAR_PADVALUE seems to be badly formatted!")
                        raise InvalidFile
                    self.skeleton.zvars[zvar]["VAR_PADVALUE"] = value
                # Retrieve value of the VAR_SPARSERECORDS parameter
                elif "!VAR_SPARSERECORDS:" in row_c:
                    value = row.strip().split(":")
                    nval = len(value)
                    if nval == 2:
                        value = value[1]
                    elif nval > 2:
                        value = ":".join(value[1:])
                    else:
                        logger.error("VAR_SPARSERECORDS seems to be badly formatted!")
                        raise InvalidFile
                    self.skeleton.zvars[zvar]["VAR_SPARSERECORDS"] = value
                elif row_c == NEW_ZVAR:
                    # New zvariable definition sub-section has been found
                    new_nrv = False
                    new_zvar = True
                    zvar = None
                    self.row = ""
                elif (
                    new_zvar
                    and row_c.startswith('"')
                    and not (row_c.endswith("F") or row_c.endswith("T"))
                ):
                    # variable definition is on several lines
                    # store these lines in self.row
                    self.row += " " + row
                elif new_zvar and (row_c.endswith("F") or row_c.endswith("T")):
                    if not self.row:
                        self.row = row
                    else:
                        self.row += " " + row
                    # Extract line containing zvariable structure (name, data type, num elem, dims, sizes
                    # rec. variance and dim variances)
                    zvar_desc = self._extract_zvar()
                    zvar = zvar_desc["Variable Name"]

                    # Initialize the zvars dictionary in Skeleton object for
                    # the given zvar
                    self.skeleton.zvars[zvar] = dict()
                    self.skeleton.zvars[zvar][NRV] = []

                    # fill the zvars dictionary in Skeleton object
                    for key in SHEETS[ZVARS]:
                        if key in zvar_desc:
                            self.skeleton.zvars[zvar][key] = zvar_desc[key]
                        else:
                            self.skeleton.zvars[zvar][key] = None
                elif new_zvar and row_c == NEW_VATTRS:
                    # New variable attribute definitions found for the current
                    # zvariable
                    new_zvar = False
                    new_vattrs = True
                    # Initialize the vattrs dictionary in Skeleton object for
                    # the given zvariable
                    self.skeleton.vattrs[zvar] = dict()
                    self.row = ""
                elif new_vattrs and row_c.startswith('"') and row_c.endswith('"-'):
                    # variable attribute definition is on several lines
                    # store these lines in self.row
                    self.row += " " + row
                elif new_vattrs and row_c.startswith('"') and row_c.endswith('"'):
                    # variable attribute definition is on several lines
                    # store these lines in self.row
                    self.row += " " + row
                elif (new_vattrs and row_c.endswith("}")) or (
                    new_vattrs and row_c.endswith("}.")
                ):
                    # End of the variable attribute definition, extract fields
                    # inside the row
                    if not self.row:
                        self.row = row
                    else:
                        self.row += " " + row
                    vatt_desc = self._extract_attr()

                    # And store them into the vattrs dictionary in Skeleton
                    # object
                    if vatt_desc["Attribute Name"] in self.skeleton.vattrs[zvar]:
                        logger.error(
                            "{0} is defined twice for {1} variable!".format(
                                vatt_desc["Attribute Name"], zvar
                            )
                        )
                    else:
                        self.skeleton.vattrs[zvar][
                            vatt_desc["Attribute Name"]
                        ] = vatt_desc
                    self.row = ""
                elif row_c == NO_NRV:
                    new_vattrs = False
                elif row_c == NEW_NRV:
                    new_vattrs = False
                    new_nrv = True
                elif new_nrv and "=" in row_c:
                    fields = self._extract_nrv(zvar, row)
                    self.skeleton.zvars[zvar][NRV].append(fields)
                # Skip comment/empty line
                elif row_c.startswith("!") or ncols == 0:
                    continue

        self.skeleton.file = self.file
        self.skeleton.xlsx = False
        return self.skeleton

    def txt_iterator(self):
        """
        Method to iter over rows of an input CDF skeleton table file.
        (rVariable parsing not supported)

        :param txt_file:
        :return: row
        """

        with open(self.file, "rt") as txt:
            for row in txt.readlines():
                logger.debug(row)

                # Read header
                if row.startswith("#" + HEADER):
                    self.section = HEADER

                # Read GLOBALattributes
                elif row.startswith("#" + GATTRS):
                    self.section = GATTRS

                # Read VARIABLEattributes
                elif row.startswith("#" + VATTRS):
                    self.section = VATTRS

                # Skip rVariables (not supported)
                elif row.startswith("#" + RVARS):
                    logger.debug("rVariables parsing not supported")
                    continue
                # Read zVariables
                elif row.startswith("#" + ZVARS):
                    self.section = ZVARS

                # End of the file
                elif row.startswith("#" + END):
                    # check that everything is correctly defined
                    if not self.skeleton.header:
                        logger.warning("{0} is missing".format(HEADER))
                    else:
                        logger.debug(HEADER + ":" + str(self.skeleton.header))
                    if not self.skeleton.gattrs:
                        logger.warning("{0} is missing".format(GATTRS))
                    else:
                        logger.debug(GATTRS + ":" + str(self.skeleton.gattrs))
                    if not self.skeleton.vattrs:
                        logger.warning("{0} is missing".format(VATTRS))
                    else:
                        logger.debug(VATTRS + ":" + str(self.skeleton.vattrs))
                    if not self.skeleton.zvars:
                        logger.warning("{0} is missing".format(ZVARS))
                    else:
                        logger.debug(ZVARS + ":" + str(self.skeleton.zvars))

                else:
                    yield row.strip()

    def _extract_attr(self):
        """
        Extract attribute from row.


        :return:
        """

        row = self.row
        # if a dot at the end of the row, then remove it
        if row.endswith("."):
            row = row[:-1]

        items = row.split()

        # Remove possible quotes at the start/end
        if items[0].startswith('"') and items[0].endswith('"'):
            attname = items[0][1:-1]
        else:
            attname = items[0]

        # If { is the first character, then it is a variable attribute
        if items[2].startswith("{"):
            entry = None
            i = 1
        # else it is a global attribute
        else:
            entry = items[1].split(":")[0]
            i = 2

        dtype = items[i]

        if dtype == "CDF_CHAR" or dtype == "CDF_UCHAR":
            att_pattern = r'\{ "([\S ]*)" \}'
        else:
            att_pattern = r"\{ ([\S ]*) \}"

        # Join value items
        value = " ".join(items[i + 1 :])

        # Remove line breaks
        value = "".join(value.split('" - "'))
        value = "".join(value.split('"- "'))
        value = "".join(value.split('" -"'))
        value = "".join(value.split('"-"'))

        # Extract attribute value from {}
        # (Use regex to make sure that it is a well formatted value)
        value = re.findall(att_pattern, value)

        if not value:
            logger.error("No value for {0}".format(attname))
            raise InvalidFile
        else:
            return dict(zip(SHEETS[GATTRS], [attname, entry, dtype, value[0]]))

    def _extract_nrv(self, zvar, row):
        """
        Extract a nrv entry

        :param row:
        :return:
        """
        items = row.strip().split("=")
        if len(items) != 2:
            logger.error("Wrong NRV for {0}".format(zvar))
            raise InvalidFile

        index = items[0].strip()[1:-1]
        value = items[1].strip()
        if value.startswith("{"):
            value = re.findall(r'\{ "([\S ]*)" \}', value)

        if isinstance(value, list):
            value = "".join(value).strip()

        return dict(zip(SHEETS[NRV], [zvar, index.strip(), value]))

    def _extract_zvar(self):
        """
        Extract a zvar definition

        :param row:
        :return:
        """
        row = self.row
        items = row.split()
        name = items[0][1:-1]
        dtype = items[1]
        elem = items[2]
        dims = items[3]
        if dims == "0":
            sizes = None
            recvar = items[4]
            dimvar = None
        else:
            i = 4
            sizes = []
            for j in range(i, i + int(dims), 1):
                sizes.append(items[j])
            i = j + 1
            recvar = items[i]
            i += 1
            dimvar = []
            for j in range(i, i + int(dims)):
                dimvar.append(items[j])

        x = [name, dtype, elem, dims, sizes, recvar, dimvar]

        return dict(zip(SHEETS[ZVARS][:8], x))


# ________________ Global Functions __________


# _________________ Main ____________________________
if __name__ == "__main__":
    print(__file__)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""Python module containing the cdfvalidator subparser."""

# ________________ IMPORT _________________________
# (Include here the modules to import, e.g. import sys)

__all__ = ["add_cdfvalidator_subparser"]

# ________________ HEADER _________________________


# ________________ Global Variables _____________
# (define here the global variables)

# ________________ Class Definition __________
# (If required, define here classes)

# ________________ Global Functions __________
# (If required, define here global functions)
def add_cdfvalidator_subparser(subparser):
    """tools.validator script program."""

    valparser = subparser.add_parser("cdf_validator", help="Validate a CDF format file")
    valparser.add_argument(
        "cdf_file",
        nargs=1,
        default=[None],
        help="Path of the CDF format file to validate",
    )
    valparser.add_argument(
        "-m",
        "--model-file",
        nargs=1,
        default=[None],
        help="Path to the model file in JSON format",
    )
    valparser.add_argument(
        "-c",
        "--cdfvalidate-bin",
        nargs=1,
        default=[None],
        help="Path of the cdfvalidate NASA CDF tool executable",
    )
    valparser.add_argument(
        "-I", "--istp", action="store_true", help="Check the ISTP guidelines compliance"
    )
    valparser.add_argument(
        "-C",
        "--run-cdfvalidate",
        action="store_true",
        help="Run the cdfvalidate NASA CDF tool",
    )

    # _________________ Main ____________________________


# if __name__ == "__main__":
#     print(__file__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ________________ HEADER _________________________


"""Skt2xlsx module

Program to convert an input CDF skeleton table file
into an Excel format file (skt_editor).

"""


# ________________ IMPORT _________________________
# (Include here the modules to import, e.g. import sys)
import os
import logging
from collections import OrderedDict
from copy import deepcopy

from openpyxl import Workbook, load_workbook

from maser.tools.toolbox import uniq
from maser.tools.cdf.serializer.exceptions import InvalidFile, InvalidEntry
from maser.tools.cdf.serializer.globals import (
    SHEETS,
    HEADER,
    GATTRS,
    ZVARS,
    VATTRS,
    NRV,
)

__all__ = ["Skt2xlsx", "Xlsx2skt"]

# ________________ Global Variables _____________
# (define here the global variables)
logger = logging.getLogger(__name__)

# ________________ Class Definition __________
# (If required, define here classes)


class Skt2xlsx:
    """Class to convert a Skeleton object into a Excel 2007 format file."""

    def __init__(self, skeleton, auto_pad=True):
        """__init__ method."""
        self.skeleton = deepcopy(skeleton)
        self.wb = None
        self.auto_pad = auto_pad

    def write_xlsx(self, output_path=None, overwrite=False):
        """
        Write the Excel 2007 format file.

        :param output_path:
        :param overwrite: If true, overwrite existing output file
        :return:
        """

        if not output_path:
            xlsx = os.path.splitext(self.skeleton.file)[0] + ".xlsx"
        elif os.path.isdir(output_path):
            xlsx = os.path.join(
                output_path,
                os.path.basename(os.path.splitext(self.skeleton.file)[0] + ".xlsx"),
            )
        else:
            xlsx = output_path

        if not overwrite and os.path.isfile(xlsx):
            logger.warning("%s already exits!", xlsx)
            return xlsx

        logger.info("Writing %s...", xlsx)

        if not self.wb:
            logger.debug("Attempting to build Excel file body...")
            self._build_wb()

        self.wb.save(filename=xlsx)

        if os.path.isfile(xlsx):
            logger.info("{0} has been saved correctly".format(xlsx))
            return xlsx
        else:
            logger.error("{0} has not been saved correctly!".format(xlsx))
            return None

    def _build_wb(self):
        """
        Create Workbook from Skeleton object.

        :return:
        """
        wb = Workbook()

        # Create header sheet
        header_sheet = wb.create_sheet(title=HEADER)

        col_idx = 1
        for key, val in self.skeleton.header.items():
            if key not in SHEETS[HEADER]:
                continue
            # Make header of header in first row
            _ = header_sheet.cell(column=col_idx, row=1, value="{0}".format(key))
            # Add values in the second row
            _ = header_sheet.cell(column=col_idx, row=2, value="{0}".format(val))
            col_idx += 1

        # Create GLOBALattributes sheet
        gattrs_sheet = wb.create_sheet(title=GATTRS)

        # Make header of GLOBALatttributes in first row
        for i, key in enumerate(SHEETS[GATTRS]):
            _ = gattrs_sheet.cell(column=i + 1, row=1, value="{0}".format(key))

        # Fill sheet with gattrs values
        row_idx = 2
        for key, entries in self.skeleton.gattrs.items():
            for i, entry in enumerate(entries):
                for j, col in enumerate(SHEETS[GATTRS]):
                    if entry[col] is None:
                        entry[col] = " "
                    _ = gattrs_sheet.cell(
                        column=j + 1, row=row_idx, value="{0}".format(entry[col])
                    )
                row_idx += 1

        # Create zVariables sheet
        zvars_sheet = wb.create_sheet(title=ZVARS)

        # Make header of zVariables in first row
        for i, key in enumerate(SHEETS[ZVARS]):
            _ = zvars_sheet.cell(column=i + 1, row=1, value="{0}".format(key))

        # Fill sheet with zvars values
        row_idx = 2
        for zvar, entries in self.skeleton.zvars.items():
            for j, col in enumerate(SHEETS[ZVARS]):
                if entries[col] is None:
                    entry = " "
                elif isinstance(entries[col], list):
                    entry = " ".join(entries[col])
                else:
                    entry = entries[col]

                _ = zvars_sheet.cell(
                    column=j + 1, row=row_idx, value="{0}".format(entry)
                )
            row_idx += 1

        # Create VARIABLEattributes sheet
        vattrs_sheet = wb.create_sheet(title=VATTRS)

        # Make header of VARIABLEattributes in first row
        for i, key in enumerate(SHEETS[VATTRS]):
            _ = vattrs_sheet.cell(column=i + 1, row=1, value="{0}".format(key))

        # Fill sheet with vattrs values
        row_idx = 2
        for zvar, vattrs in self.skeleton.vattrs.items():
            for vatt, entry in vattrs.items():
                # First column contains Variable name
                _ = vattrs_sheet.cell(column=1, row=row_idx, value="{0}".format(zvar))
                for j, col in enumerate(SHEETS[VATTRS][1:]):
                    if entry[col] is None:
                        entry[col] = " "
                    _ = vattrs_sheet.cell(
                        column=j + 2, row=row_idx, value="{0}".format(entry[col])
                    )
                row_idx += 1

        # Create NRV sheet
        nrv_sheet = wb.create_sheet(title=NRV)

        # Make header of NRV in first row
        for i, key in enumerate(SHEETS[NRV]):
            _ = nrv_sheet.cell(column=i + 1, row=1, value="{0}".format(key))

        row_idx = 2
        for zvar, fields in self.skeleton.zvars.items():
            if "NRV" in fields:
                for i, entries in enumerate(fields["NRV"]):
                    for j, col in enumerate(SHEETS[NRV]):
                        if entries[col] is None:
                            entry[col] = " "
                        elif isinstance(entries[col], list):
                            entry = " ".join(entries[col])
                        else:
                            entry = entries[col]

                        _ = nrv_sheet.cell(
                            column=j + 1, row=row_idx, value="{0}".format(entry)
                        )
                    row_idx += 1

        self.wb = wb


# ________________ Global Functions __________


class Xlsx2skt:
    """Class to convert an Excel 2007 format file into a Skeleton object."""

    def __init__(self, skeleton):

        self.file = None
        self.skeleton = skeleton
        self.cdf_items = dict()

    def parse_xlsx(self, xlsx_file, auto_pad=True):
        """
        Parse the Excel 2007 format file.

        :param: Input Excel 2007 format file
        :return:
        """
        self.file = xlsx_file
        if not os.path.isfile(xlsx_file):
            logger.error("Cannot find Excel file called %s!", xlsx_file)
            raise FileNotFoundError

        if os.path.splitext(xlsx_file)[1] != ".xlsx":
            logger.error("Invalid input Excel format!")
            raise InvalidFile

        logger.info("Parsing %s file...", xlsx_file)
        wkbk = load_workbook(xlsx_file, read_only=True)
        sheet_names = wkbk.sheetnames

        if ("rVariables" in sheet_names) or ("variables" in sheet_names):
            logger.warning("rVariable type is not supported!")

        sheets = dict()
        for shtn in SHEETS:
            logger.debug("Loading %s sheet...", shtn)
            if shtn not in sheet_names:
                logger.error("Missing %s sheet in the input Excel file!", shtn)
                raise InvalidFile
            else:
                wksht = wkbk[shtn]

                sheet_data = OrderedDict()
                for i, row in enumerate(wksht.rows):
                    # Get sheet columns names on the first row
                    # And initialize columns lists
                    if i == 0:
                        header = []
                        for cell in row:
                            if cell.value is None:
                                continue
                            sheet_data[cell.value] = []
                            header.append(cell.value)

                        # Check that file header contains expected columns
                        for col in SHEETS[shtn]:
                            if col not in header:
                                logger.error(
                                    "Missing %s column in the input Excel file!", col
                                )
                                raise InvalidFile

                    # Then, for other rows get cell values for each column
                    else:
                        ncell = len(row)
                        for j, key in enumerate(header):
                            if ncell > j:
                                row_j = str(row[j].value)
                                if row_j.startswith('"') and row_j.endswith('"'):
                                    row_j = row_j[1:-1]
                                else:
                                    row_j = row[j].value
                                sheet_data[key].append(row_j)
                            else:
                                logger.debug("Warning -- empty cell!")
                                sheet_data[key].append(None)

                if not sheet_data:
                    raise InvalidFile(
                        "{0} is invalid: "
                        "{1} sheet has no header nor data!".format(xlsx_file, shtn)
                    )

                sheets[shtn] = sheet_data

        self.cdf_items[GATTRS] = uniq(sheets[GATTRS]["Attribute Name"], not_none=True)
        self.cdf_items[VATTRS] = uniq(sheets[VATTRS]["Attribute Name"], not_none=True)
        self.cdf_items[ZVARS] = uniq(sheets[ZVARS]["Variable Name"], not_none=True)
        logger.debug("%i GLOBAL attributes returned", len(self.cdf_items[GATTRS]))
        logger.debug("%i Variable attributes returned", len(self.cdf_items[VATTRS]))
        logger.debug("%i zVariables returned", len(self.cdf_items[ZVARS]))

        self._xlsx2skt(sheets, auto_pad=auto_pad)

        return self.skeleton

    def _xlsx2skt(self, sheets, auto_pad=True):
        """
        Store input Excel sheets into Skeleton attributes

        :param sheets:
        :return:
        """

        header = dict()
        gattrs = dict()
        vattrs = dict()
        zvars = dict()

        # Fill header
        try:
            for key, value in sheets[HEADER].items():
                header[key] = value[0]
        except Exception:
            logger.exception(
                "{0} sheet is badly formatted, please check!".format(HEADER)
            )
            raise InvalidFile()

        # Add items number list
        header["nzvar"] = len(self.cdf_items[ZVARS])
        header["ngattr"] = len(self.cdf_items[GATTRS])
        header["nvattr"] = len(self.cdf_items[VATTRS])

        # Fill gattrs
        try:
            for gatt in self.cdf_items[GATTRS]:
                gattrs[gatt] = []
                for i, name in enumerate(sheets[GATTRS]["Attribute Name"]):
                    if gatt == name:
                        fields = {}
                        for col in SHEETS[GATTRS]:
                            current_value = sheets[GATTRS][col][i]
                            # If NoneType or empty string, set to one space string
                            if (
                                current_value is None
                                or str(current_value).strip() == ""
                            ):
                                current_value = " "
                            fields[col] = current_value

                        gattrs[gatt].append(fields)
        except Exception:
            raise InvalidFile(
                message="{0} sheet is badly formatted, please check!".format(GATTRS)
            )

        # Fill zvars
        try:
            for i, zvar in enumerate(sheets[ZVARS]["Variable Name"]):
                if zvar is None:
                    continue

                fields = dict()
                for col in SHEETS[ZVARS]:
                    try:
                        if sheets[ZVARS][col][i] is None:
                            current_value = ""
                        else:
                            current_value = str(sheets[ZVARS][col][i]).strip()
                        # Fill column values for current zvar
                        if col == "Sizes":
                            # If Sizes, make sure to have vector
                            fields[col] = [current_value]
                        else:
                            fields[col] = current_value
                    except Exception:
                        raise InvalidEntry(
                            message="Invalid {0} for {1} zVariable!".format(col, zvar)
                        )

                if auto_pad:
                    fields["VAR_PADVALUE"] = assign_pad(sheets[ZVARS]["Data Type"][i])

                zvars[zvar] = fields

                # Add NRV for this zvar (if exist, if not set to empty list [])
                zvars[zvar][NRV] = []
                for i, name in enumerate(sheets[NRV]["Variable Name"]):
                    if zvar == name:
                        fields = dict()
                        for col in SHEETS[NRV]:
                            fields[col] = sheets[NRV][col][i]
                        zvars[zvar][NRV].append(fields)

                # Fill variable attributes for this zvar
                for j, name in enumerate(sheets[VATTRS]["Variable Name"]):
                    if zvar == name:
                        if zvar not in vattrs:
                            vattrs[zvar] = dict()
                        fields = dict()
                        for col in SHEETS[VATTRS][1:]:
                            fields[col] = sheets[VATTRS][col][j]

                        vattrs[zvar][sheets[VATTRS]["Attribute Name"][j]] = fields
        except Exception:
            logger.exception(
                "{0}, {1} or {2} sheet is badly formatted, "
                "please check!".format(ZVARS, VATTRS, NRV)
            )
            raise InvalidFile(message="{0} is invalid!".format(self.file))

        # Store results into self.skeleton
        self.skeleton.header = header
        self.skeleton.gattrs = gattrs
        self.skeleton.zvars = zvars
        self.skeleton.vattrs = vattrs
        self.skeleton.file = self.file
        self.skeleton.xlsx = True
        self.skeleton.cdf_items = self.cdf_items
        self.skeleton.vattrList = self.cdf_items[VATTRS]


# ________________ Global Functions __________


def assign_pad(data_type):
    """VAR_PADVALUE auto assign.

    Automatically assigns VAR_PADVALUE
    depending to the input data_type
    """
    dtype = data_type.upper()

    if "EPOCH" in dtype:
        return "01-Jan-0000 00:00:00.000"
    elif "TT2000" in dtype:
        return "0000-01-01T00:00:00.000000000"
    elif ("INT" in dtype) or ("BYTE" in dtype):
        return "0"
    elif ("FLOAT" in dtype) or ("REAL" in dtype) or ("DOUBLE" in dtype):
        return "0.0"
    elif "CHAR" in dtype:
        return '" "'
    else:
        return "None"


# _________________ Main ____________________________
if __name__ == "__main__":
    print(__file__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from maser.tools.cdf.serializer import skeletoncdf  # noqa: F401
from maser.tools.cdf.validator import Validate  # noqa: F401
from maser.tools.cdf.tools import (  # noqa: F401
    get_cdftype,
    get_vattrs,
    get_cdftypename,
    get_numpttype,
)
from maser.tools.cdf.cdfcompare.cdf_compare import cdf_compare  # noqa: F401

# from maser.tools.settings import ROOT_DIRECTORY, MASER_LIB_DIR, SUPPORT_DIR, __version__



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
leapsec.py.

Python module to load and handle the
NASA CDF LeapSecond table.


"""

# ________________ IMPORT _________________________
# (Include here the modules to import, e.g. import sys)
import os
import logging
from datetime import datetime

from maser.tools.toolbox import download_data, print_exception
from maser.tools.settings import SUPPORT_DIR

__all__ = ["Lstable"]

# ________________ HEADER _________________________


# ________________ Global Variables _____________
# (define here the global variables)
logger = logging.getLogger(__name__)

URL = "https://cdf.gsfc.nasa.gov/html/CDFLeapSeconds.txt"
ENVAR = "CDF_LEAPSECONDSTABLE"

LS_FILENAME = "CDFLeapSeconds.txt"

INPUT_DATE = "%Y-%m-%dT%H:%M:%S"

LS_FILE_DEF_DIR = os.path.join(SUPPORT_DIR, "data")
if not os.path.isdir(LS_FILE_DEF_DIR):
    os.makedirs(LS_FILE_DEF_DIR)
LS_FILE_DEF_PATH = os.path.join(LS_FILE_DEF_DIR, LS_FILENAME)

# ________________ Class Definition __________
# (If required, define here classes)


class LstableException(Exception):
    pass


class Lstable:
    """Class for the leapsec table."""

    def __init__(self, file=None):
        """Leapsec __init__ method."""
        self.file = None
        self.date = []
        self.leapsec = []
        self.drift = []
        self.lstable = None

        # Setting the CDFLeapSeconds.txt filepath
        # and loading table
        self.set_file(file=file, reload=True)

    def _add(self, row):
        """Add a new set of data to the leapsec table."""
        fields = row.split()
        if len(fields) != 6:
            logger.error("Input row is not valid!")
            return

        date = datetime(int(fields[0]), int(fields[1]), int(fields[2]))
        self.date.append(date)
        self.leapsec.append(float(fields[3]))
        self.drift.append([float(fields[4]), float(fields[5])])

    def get_leapsec(self, date=datetime.now()):
        """Return the leapseconds for a given datetime."""
        if self.lstable is None:
            self.load_lstable()

        if date < self.date[0]:
            return 0.0
        elif date >= self.date[-1]:
            return self.leapsec[-1]
        else:
            for i, lsdate in enumerate(self.date[:-1]):
                if date >= lsdate and date < self.date[i + 1]:
                    return self.leapsec[i]

        return None

    @staticmethod
    def get_lstable_file(target_dir=None, overwrite=False, url=URL):
        """Download the CDFLeapSeconds.txt leapsec table file into the target_dir.

        If target_dir is None and the $CDF_LEAPSECONDSTABLE env. variable
        is defined, then the file is saved into $CDF_LEAPSECONDSTABLE path.
        If target_dir is None and the $CDF_LEAPSECONDSTABLE env. variable
        is not defined, then file is copied in LS_FILE_DEF_PATH default path.

        Default file url is https://cdf.gsfc.nasa.gov/html/CDFLeapSeconds.txt.
        However, it can be changed using the url input keyword.

        overwrite input keyword allows program to replace existing file
        """
        if target_dir is None and ENVAR in os.environ:
            target_filepath = os.environ[ENVAR]
        elif target_dir is None and ENVAR not in os.environ:
            target_filepath = LS_FILE_DEF_PATH
        elif target_dir is not None:
            target_filepath = os.path.join(target_dir, LS_FILENAME)

        if os.path.isfile(target_filepath):
            logger.info("{0} already exists!".format(target_filepath))
            if overwrite:
                logger.warning("{0} will be replaced!".format(target_filepath))
                os.remove(target_filepath)
            else:
                return False

        try:
            if download_data(url, target_file=target_filepath) is not None:
                logger.info("{0} saved".format(target_filepath))
            else:
                raise LstableException("Downloading {0} has failed!".format(url))
        except Exception:
            print_exception()
        else:
            return True

    def set_file(self, file=None, reload=True):
        """Check if the leapsec table file exists on the local disk."""
        if file is not None:
            self.file = file
        elif file is None and ENVAR in os.environ:
            self.file = os.environ[ENVAR]
        elif (
            file is None
            and ENVAR not in os.environ
            and os.path.isfile(LS_FILE_DEF_PATH)
        ):
            self.file = LS_FILE_DEF_PATH
        else:
            self.file = URL

        # Reloading the lstable with file
        if reload:
            self.load_lstable(reload=True)

    def load_lstable(self, reload=False):
        """Load NASA CDF leapsec table CDFLeapSeconds.txt."""
        if self.lstable is not None and not reload:
            logger.info("Leapsec. table is already loaded, aborting!")
            return False
        elif self.lstable is not None and reload:
            logger.warning("Leapsec. table will be reloaded!")

        if self.file.startswith("http") or self.file.startswith("ftp"):
            data = download_data(self.file)
            self.lstable = self._parse_lstable(data)
            return True
        elif os.path.isfile(self.file):
            buff = open(self.file, "rt")
            data = buff.read()
            self.lstable = self._parse_lstable(data)
            return True
        else:
            logger.error(
                "CDFLeapSeconds.txt cannot be loaded! Make sure file exists locally or download it first using command 'maser leapsec -D'"
            )
            return False

    def _parse_lstable(self, data):
        """Parse the CDF leap second table file."""
        for row in data.split("\n"):
            row = str(row).rstrip()
            if row.startswith(";"):
                continue
            else:
                if row.strip() == "":
                    continue
                self._add(row)

        return data

    def __str__(self):
        """__str__ method."""
        if self.file is None:
            self.set_file()

        self.load_lstable()
        if self.lstable is None:
            return ""
        string = "Date -- Leap Sec. -- Drift\n"
        for i, row in enumerate(self.date):
            string += "{0} -- {1} -- {2}\n".format(
                self.date[i], self.leapsec[i], self.drift[i]
            )
        return string


# ________________ Global Functions __________
# (If required, define here gobal functions)


# _________________ Main ____________________________
# if (__name__ == "__main__"):
#    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""Contains unit tests of maser4py time module."""

# ________________ IMPORT _________________________
# (Include here the modules to import, e.g. import sys)
from datetime import datetime, timedelta

from numpy import datetime64, timedelta64

from .time import (
    get_leapsec,
    local_to_utc,
    tt2000_to_utc,
    utc_to_tt2000,
    tt2000_to_jd,
    jd_to_tt2000,
)

# ________________ HEADER _________________________

# Mandatory
# __version__ = ""
# __author__ = ""
# __date__ = ""

# # Optional
# __license__ = ""
# __credit__ = [""]
# __maintainer__ = ""
# __email__ = ""
# __project__ = ""
# __institute__ = ""
# __changes__ = ""


# ________________ Global Variables _____________
# (define here the global variables)

# ________________ Class Definition __________
# (If required, define here classes)

# ________________ Global Functions __________
# (If required, define here gobal functions)

# Test time.py methods


def test_get_leapsec():
    """Test get_leapsec()."""
    date = datetime64("2011-03-01T07:00")
    leapsec = get_leapsec(date)
    assert leapsec == timedelta(seconds=34)


def test_get_leapsec64():
    """Test get_leapsec64().

    Same than test_get_leapsec(), but
    with leapsec returned as timedelta64
    """
    date = datetime64("2011-03-01T07:00")
    leapsec = get_leapsec(date, to_timedelta64=True)
    assert leapsec == timedelta64(34, "s")


def test_local_to_utc():
    """Test local_to_utc()."""
    utc = local_to_utc(datetime64("2011-03-01T07:00"), tzone="Europe/Paris")
    assert utc == datetime64("2011-03-01T06:00")


def test_utc_to_tt2000():
    """Test utc_to_tt2000()."""
    tt2000 = utc_to_tt2000(datetime(2011, 3, 1, 7, 0, 0))
    assert tt2000 == timedelta64(352234802000000000, "ns")


def test_tt2000_to_utc():
    """Test tt2000_to_utc()."""
    tt2000 = timedelta64(352234802000000000, "ns")
    utc = tt2000_to_utc(tt2000)
    assert utc == datetime64("2011-03-01T07:00")


def test_tt2000_to_jd():
    """Test tt2000_to_jd()."""
    tt2000 = timedelta64(352234802000000000, "ns")
    jd = tt2000_to_jd(tt2000)
    assert jd == timedelta64(212165722800000000, "us")


def test_mjd_to_tt2000():
    """Test mjd_to_tt2000()."""
    mjd = timedelta64(80094660, "m")
    tt2000 = jd_to_tt2000(mjd, from_mjd=True)
    assert tt2000 == timedelta64(352234802000000, "us")


# _________________ Main ____________________________
# if (__name__ == "__main__"):
# print ""
# main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module to install a given CDF software release.
@Sonny Lion (LESIA)
"""

import argparse
import os
import platform
import subprocess
import tarfile
from io import BytesIO

import requests

DEFAULT_CDF_URL = "http://cdaweb.gsfc.nasa.gov/pub/software/cdf/dist/cdf{cdf_version}/linux/cdf{cdf_version}-dist-cdf.tar.gz"


def get_cdf(raw_cdf_version, install_dir, cdf_url=DEFAULT_CDF_URL):
    # reformat the cdf version from X.Y.Z to XY_Z
    X, Y, Z = raw_cdf_version.split(".")
    cdf_version = f"{X}{Y}_{Z}"

    # build the url with the cdf version
    url = cdf_url.format(cdf_version=cdf_version)

    response = requests.get(url)
    with tarfile.open(fileobj=BytesIO(response.content), mode="r:gz") as tar:
        tar.extractall(path=install_dir)


def make_cdf(raw_cdf_version, install_dir):
    # reformat the cdf version from X.Y.Z to XY_Z
    X, Y, Z = raw_cdf_version.split(".")
    cdf_version = f"{X}{Y}_{Z}"

    # prepare the make command
    command = ["make", "-C", os.path.join(install_dir, f"cdf{cdf_version}-dist")]
    if platform.system() == "Linux":
        command += ["OS=linux", "ENV=gnu", "CURSES=yes", "all"]
    elif platform.system() == "Darwin":
        command += ["OS=macosx", "ENV=x86_64", "CURSES=yes", "all"]
    else:
        raise NotImplementedError(
            "Your OS is not supported at the moment, please install the CDF lib manually"
        )

    # run the command
    subprocess.check_call(command)


def install_cdf(raw_cdf_version, install_dir):
    # reformat the cdf version from X.Y.Z to XY_Z
    X, Y, Z = raw_cdf_version.split(".")
    cdf_version = f"{X}{Y}_{Z}"

    current_dir = os.path.abspath(os.path.join(install_dir, "current"))

    # prepare the make command
    command = [
        "make",
        "-C",
        os.path.join(install_dir, f"cdf{cdf_version}-dist"),
        f"INSTALLDIR={current_dir}",
        "install",
    ]

    # run the command
    subprocess.check_call(command)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("cdf_version", help="Version of the CDF to retrieve.")
    parser.add_argument(
        "-i",
        "--install-dir",
        help="Local directory where IDB source files will be saved",
    )
    parser.add_argument(
        "-c",
        "--cdf-url",
        help="URL or local path of the NASA CDF dist. source tar.gz file",
        default=default_cdf_url,
    )
    parser.add_argument(
        "--make",
        action="store_true",
        default=False,
        help="If passed, then build and install the CDF software",
    )

    args = parser.parse_args()

    get_cdf(args.cdf_version, args.install_dir, args.cdf_url)

    if args.make:
        make_cdf(args.cdf_version, args.install_dir)
        install_cdf(args.cdf_version, args.install_dir)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from maser.tools.cdf.serializer.exceptions import *  # noqa: F401, F403
from maser.tools.cdf.serializer.subparser import *  # noqa: F401, F403
from maser.tools.cdf.serializer.skeletoncdf import *  # noqa: F401, F403
from maser.tools.cdf.serializer.skeletontable import *  # noqa: F401, F403
from maser.tools.cdf.serializer.skeleton import *  # noqa: F401, F403



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""Python module template."""

# ________________ IMPORT _________________________
# (Include here the modules to import, e.g. import sys)
import os

from maser.tools.settings import SUPPORT_DIR

# ________________ HEADER _________________________


# ________________ Global Variables _____________
# (define here the global variables)
LS_FILENAME = "CDFLeapSeconds.txt"

CURDIR = os.path.dirname(os.path.abspath(__file__))
LS_FILE_DEF_DIR = os.path.join(SUPPORT_DIR, "data")
if not os.path.isdir(LS_FILE_DEF_DIR):
    os.makedirs(LS_FILE_DEF_DIR)
LS_FILE_DEF_PATH = os.path.join(LS_FILE_DEF_DIR, LS_FILENAME)

# ________________ Class Definition __________
# (If required, define here classes)


# ________________ Global Functions __________
# (If required, define here global functions)
def add_leapsec_subparser(subparser):
    """tools.serializer script program."""
    leapsec_parser = subparser.add_parser("leapsec", help="Leapsecond handling tool")
    leapsec_parser.add_argument(
        "-f",
        "--filepath",
        nargs=1,
        default=[LS_FILE_DEF_PATH],
        help="CDFLeapSeconds.txt filepath.\n "
        "Default is ${0}".format(LS_FILE_DEF_PATH),
    )
    leapsec_parser.add_argument(
        "-d",
        "--date",
        nargs=1,
        default=[None],
        help="Return the leap seconds for "
        "a given date and time."
        '(Expected format is "YYYY-MM-DDThh:mm:ss")',
    )
    leapsec_parser.add_argument(
        "-D",
        "--DOWNLOAD-FILE",
        action="store_true",
        help="Download the CDFLeapSeconds.txt file"
        "from the NASA CDF Web site. "
        "The file will be saved in the path"
        " defined in the --filepath argument.",
    )
    leapsec_parser.add_argument(
        "-S", "--SHOW-TABLE", action="store_true", help="Show the leap sec. table"
    )
    leapsec_parser.add_argument(
        "-O", "--OVERWRITE", action="store_true", help="Overwrite existing file"
    )

    # _________________ Main ____________________________


# if __name__ == "__main__":
#     print(__file__)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['maser',
 'maser.tools',
 'maser.tools.cdf',
 'maser.tools.cdf.cdfcompare',
 'maser.tools.cdf.serializer',
 'maser.tools.cdf.validator',
 'maser.tools.time']

package_data = \
{'': ['*'], 'maser.tools': ['support/cdf/*', 'support/data/*']}

install_requires = \
['numpy>=1.23.3,<2.0.0',
 'openpyxl>=3.0.10,<4.0.0',
 'pytz>=2022.4,<2023.0',
 'spacepy>=0.4.0,<0.5.0']

setup_kwargs = {
    'name': 'maser-tools',
    'version': '0.1.2',
    'description': '',
    'long_description': 'None',
    'author': 'MASER Team',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""Toolbox module for maser-py package."""

# ________________ IMPORT _________________________
# (Include here the modules to import, e.g. import sys)
import sys
import os
import os.path as osp
import re
import logging
import subprocess
import urllib.request
import traceback
import shutil
import errno

import tempfile

__all__ = [
    "get_version",
    "print_exception",
    "which",
    "truncate_str",
    "Singleton",
    "download_data",
    "setup_logging",
    "quote",
    "uniq",
    "insert_char",
    "run_command",
    "move_safe",
]

# ________________ HEADER _________________________

# # Mandatory
# __version__ = ""
# __author__ = ""
# __date__ = ""
#
# # Optional
# __institute__ = ""
# __project__ = ""
# __license__ = ""
# __credit__ = [""]
# __maintainer__ = ""
# __email__ = ""
# __change__ = {"version": "change"}

# ________________ Global Variables _____________
# (define here the global variables)
LOGGER = logging.getLogger(__name__)

DEF_INDENT = " " * 16

# ________________ Class Definition __________
# (If required, define here classes)


class Singleton(type):
    """
    Singleton class.

    A metaclass to create singletons, i.e classes that can have at most only
    one instance created at a given time.
    """

    def __call__(cls, *args, **kwargs):
        """
        Singleton.__call__.

        Check that an instance is already
        stored before creating a new one.
        """
        if hasattr(cls, "instance"):
            return cls.instance

        cls.instance = super(Singleton, cls).__call__(*args, **kwargs)

        return cls.instance


# ________________ Global Functions __________
# (If required, define here global functions)


def get_version(changelog):
    """Get latest version from the input CHANGELOG.md file."""
    pattern = re.compile(r"(\d*)\.(\w*)\.(\w*)")
    if osp.isfile(changelog):
        with open(changelog, "rt") as file:
            for line in file:
                if pattern.match(line):
                    return line.strip()

    print("WARNING: CHANGELOG.md not found or invalid, version unknown!")
    return "unknown"


def move_safe(src, dst, no_erase=False, tempdir=tempfile.gettempdir()):
    """
    Safely move a sections file or a directory into a dst directory.

    If no_erase keyword is True, then do no delete the original file/folder
    but move it to a tempdir in case of.
    """
    # Build target file/folder
    target = os.path.join(dst, os.path.basename(src))

    # copy file/folder
    try:
        shutil.copytree(src, target)
        is_item = os.path.isdir
    except shutil.Error as e:
        LOGGER.error(f"Directory not copied. Error: {e}")
    except OSError as e:
        if e.errno == errno.ENOTDIR:
            shutil.copyfile(src, dst)
            is_item = os.path.isfile
        else:
            LOGGER.error("{0} is not a valid directory!".format(dst))
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), src)

    # Check that file/directory has been correctly copied before removing the
    # original
    if not is_item(target):
        LOGGER.error("{0} has not been moved correctly, aborting!".format(src))
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), src)

    if no_erase:
        move_safe(src, tempdir)
        LOGGER.info("{0} moved in {1}".format(src, tempdir))

    return target


def download_data(url, encoding="utf-8", target_file=None, binary=False):
    """Download the data from the given url.

    If target_file input keyword is not None, then
    save data into the target_file path.
    """
    try:
        buff = urllib.request.urlopen(url)
        data = buff.read().decode(encoding)
    except Exception:
        return None
    else:
        if target_file is not None:
            if binary:
                wm = "wb"
            else:
                wm = "w"
            with open(target_file, wm) as fw:
                fw.write(data)
    return data


def insert_char(string, char, pos):
    """Insert substring in a string."""
    return string[:pos] + char + string[pos:]


def print_exception(message=None, exit=True):
    """
    Print_exception.

    To handle the printing of the error message when one occurred. Particularly
    useful when catching errors and want to add debug informations on the same
    time.
    """
    # get the traceback
    trace = traceback.format_exc()

    # if not message provided, get the traceback of errors to be a little more
    # useful for the developer
    if message is not None:
        mess = "\n".join([trace, message])
    else:
        # else use message provided by developer
        mess = trace

    # show error in the logger
    LOGGER.error(mess)

    if exit:
        sys.exit(1)

    # return the message
    return mess


def quote(string, unquote=False):
    """Double quote a given string."""
    if string is not None:
        if not isinstance(string, str):
            string = str(string)
        if string.startswith('"'):
            string = string[1:]
        if string.endswith('"'):
            string = string[:-1]
        if unquote:
            return string
        return '"' + string + '"'


def run_command(
    cmd,
    env=None,
    stdout=subprocess.PIPE,
    stderr=subprocess.PIPE,
    shell=False,
    logger=LOGGER,
):
    """Run a command with subprocess."""

    try:
        logger.info(" ".join(cmd))
        res = subprocess.Popen(cmd, env=env, shell=shell, stdout=stdout, stderr=stderr)
    except TypeError as e:
        logger.error(e)
    except OSError as e:
        logger.error(e)
    except subprocess.TimeoutExpired as e:
        logger.error("TIME OUT EXPIRED:  %i SEC.", e.timeout)

    return res


def set_level(logger_or_handler, quiet=False, debug=False):
    if debug:
        logger_or_handler.setLevel(logging.DEBUG)
    elif quiet:
        logger_or_handler.setLevel(logging.CRITICAL + 10)
    else:
        logger_or_handler.setLevel(logging.INFO)


def set_handler_config(handler, quiet=False, debug=False, formatter=None):
    set_level(handler, quiet=quiet, debug=debug)
    if formatter:
        handler.setFormatter(formatter)


def setup_logging(
    filename=None, quiet=False, debug=False, logger=logging.root, stream=None
):
    """Method to set up logging."""

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S",
    )

    # set the logger log-level
    set_level(logger, quiet=quiet, debug=debug)

    # if no handler is provided
    if stream is None:
        # and no handler already exists
        if not logger.handlers:
            # set a default stream handler
            logger.addHandler(logging.StreamHandler())
    else:
        # add the given stream handler
        logger.addHandler(stream)

    # loop over existing handlers and set the config
    for handler in logger.handlers:
        set_handler_config(handler, quiet=quiet, debug=debug, formatter=formatter)

    # create the file handler
    if filename:
        fh = logging.FileHandler(filename, delay=True)
        fh.setFormatter(
            logging.Formatter(
                "%(asctime)s %(name)-\
                        12s %(levelname)-8s %(funcName)-12s %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        )
        if debug:
            fh.setLevel(logging.DEBUG)
        else:
            fh.setLevel(logging.INFO)

        logger.addHandler(fh)

    return logger


def truncate_str(string, max_length, gap=DEF_INDENT, min_length=3):
    """Truncate a too long CDF_CHAR value."""
    nstr = len(string)
    new_string = ""
    for i, val_c in enumerate(string):
        if i > nstr - min_length:
            new_string += string[i:]
            break
        new_string += val_c
        if (i % max_length == 0) and (i != 0):
            new_string += '" - \n' + gap + '"'

    return new_string


def uniq(seq, not_none=False):
    """Get list of unique elements from an input sequence of list type."""
    seen = set()
    seen_add = seen.add
    if not_none:
        return [x for x in seq if not (x in seen or seen_add(x) or x is None)]
    else:
        return [x for x in seq if not (x in seen or seen_add(x))]


def which(program, path="PATH"):
    """Run which function."""

    def is_exe(fpath):
        """Run is_exe function."""
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, __ = os.path.split(program)

    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ[path].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


# _________________ Main ____________________________
if __name__ == "__main__":
    print("toolbox for maser package")
